#cython: language_level=3

from pathlib import Path

from obitools3.utils cimport bytes2str, str2bytes

from .capi.obidms cimport obi_dms, \
                          obi_close_dms
from .capi.obidmscolumn cimport obi_column_get_header_from_name, \
                                obi_unmap_header, \
                                obi_column_get_latest_version_from_name, \
                                obi_create_column, \
                                obi_clone_column, \
                                obi_open_column, \
                                obi_close_column, \
                                obi_column_format_date, \
                                OBIDMS_column_header_p
from .capi.obitypes cimport const_char_p, \
                            name_data_type
                            

from ._obidms cimport OBIDMS
from ._obidms cimport OBIDMS_column

from ._obidmscolumn_int cimport OBIDMS_column_int, \
                                OBIDMS_column_int_writable, \
                                OBIDMS_column_int_multi_elts, \
                                OBIDMS_column_int_multi_elts_writable
                                
from ._obidmscolumn_float cimport OBIDMS_column_float, \
                                  OBIDMS_column_float_writable, \
                                  OBIDMS_column_float_multi_elts, \
                                  OBIDMS_column_float_multi_elts_writable

from ._obidmscolumn_bool cimport OBIDMS_column_bool, \
                                 OBIDMS_column_bool_writable, \
                                 OBIDMS_column_bool_multi_elts, \
                                 OBIDMS_column_bool_multi_elts_writable

from ._obidmscolumn_char cimport OBIDMS_column_char, \
                                 OBIDMS_column_char_writable, \
                                 OBIDMS_column_char_multi_elts, \
                                 OBIDMS_column_char_multi_elts_writable

from ._obidmscolumn_str cimport OBIDMS_column_str, \
                                OBIDMS_column_str_writable, \
                                OBIDMS_column_str_multi_elts, \
                                OBIDMS_column_str_multi_elts_writable


cdef class OBIDMS :
    
    def __init__(self, str dms_name) :
        
        # Declarations
        cdef bytes dms_name_b
        
        # Format the character string to send to C function
        dms_name_b = str2bytes(dms_name)
        
        # Fill structure and create or open the DMS
        self.dms_name = dms_name
        self.pointer = obi_dms(<const_char_p> dms_name_b)
        # TODO: test pointer and raise Exception("Failed opening or creating an OBIDMS")


    cpdef close(self) :
        #TODO close all columns
        if (obi_close_dms(self.pointer)) < 0 :
            raise Exception("Problem closing an OBIDMS")


    cpdef dict list(self):
    
        # Declarations
        cdef object p
        cdef dict dms = {}
        cdef str column_name
        cdef bytes column_name_b
        cdef str data_type
        cdef str creation_date
        cdef obiversion_t latest_version
        cdef index_t line_count
        cdef OBIDMS_column_header_p header
        
        p = Path(self.dms_name+'.obidms')

        print("{:<30} {:<12} {:<25} {:<30} {:<40}".format('-Column name-', 
                                                   '-Data type-', 
                                                   '-Latest version number-', 
                                                   '-Line count of latest version-',
                                                   '-Creation date of latest version-'))
        for entry in p.iterdir():
            if entry.suffix == ".obicol":
                column_name = entry.stem
                column_name_b = str2bytes(column_name)
                dms[column_name] = {}
                header = obi_column_get_header_from_name(self.pointer, column_name_b)
                data_type = bytes2str(name_data_type(header.data_type))
                line_count = header.line_count
                creation_date = bytes2str(obi_column_format_date(header.creation_date))
                obi_unmap_header(header)    # TODO check if error? but C will already warn and there's nothing to do
                latest_version = obi_column_get_latest_version_from_name(self.pointer, column_name_b)
                dms[column_name]['data_type'] = data_type
                dms[column_name]['latest_version'] = latest_version
                dms[column_name]['line_count'] = line_count
                dms[column_name]['creation_date'] = creation_date
                # TODO : actually get all the informations in the header
                print("{:<30} {:<12} {:<25} {:<30} {:<40}".format(column_name, data_type, latest_version, line_count, creation_date))

        return dms
     
     
    cpdef OBIDMS_column open_column(self,
                                    str column_name,
                                    bint create=False, 
                                    bint clone=False, bint clone_data=True,
                                    obiversion_t version_number=-1, 
                                    OBIType_t data_type= <OBIType_t> 0,
                                    index_t nb_lines=0, 
                                    index_t nb_elements_per_line=0, 
                                    list elements_names=None,
                                    str array_name="default_obiarray"):
        
        # Declarations
        cdef OBIDMS_column column
        cdef object subclass        # TODO object?
        cdef bytes column_name_b
        cdef OBIDMS_column_header_p header
        
        header = NULL
        
        # Format the character string to send to C function
        column_name_b = str2bytes(column_name)
        
        # Get the header of the latest version of the column if 
        # some needed informations are not provided
        if ((not data_type or not nb_elements_per_line) and not create) :
            header = obi_column_get_header_from_name(self.pointer, column_name_b)
            
        # Get the data type if not provided
        if not data_type :
            if create :
                raise Exception("A data type must be specified")
            else :
                data_type = header.data_type

        # Get the number of elements per line if not provided and needed
        if not nb_elements_per_line :
            if create : # Set to one if not provided (default value)
                nb_elements_per_line = 1
            else :
                nb_elements_per_line = header.nb_elements_per_line
                if nb_elements_per_line > 1 :
                    elements_names = bytes2str(header.elements_names).split(';')

        if header != NULL :
            obi_unmap_header(header)        # TODO check if error? but C will already warn and there's nothing to do
        
        # Open the column with the right subclass depending on the data type, the mode 
        # (read-only or writable) and whether there are multiple elements per line or not  
        if data_type == 1 :
            if (create or clone) :
                if nb_elements_per_line == 1 :
                    subclass = OBIDMS_column_int_writable
                else :
                    subclass = OBIDMS_column_int_multi_elts_writable
            else :
                if nb_elements_per_line == 1 :
                    subclass = OBIDMS_column_int
                else :
                    subclass = OBIDMS_column_int_multi_elts
        elif data_type == 2 :
            if (create or clone) :
                if nb_elements_per_line == 1 :
                    subclass = OBIDMS_column_float_writable
                else :
                    subclass = OBIDMS_column_float_multi_elts_writable
            else :
                if nb_elements_per_line == 1 :
                    subclass = OBIDMS_column_float
                else :
                    subclass = OBIDMS_column_float_multi_elts
        elif data_type == 3 :
            if (create or clone) :
                if nb_elements_per_line == 1 :
                    subclass = OBIDMS_column_bool_writable
                else :
                    subclass = OBIDMS_column_bool_multi_elts_writable
            else :
                if nb_elements_per_line == 1 :
                    subclass = OBIDMS_column_bool
                else :
                    subclass = OBIDMS_column_bool_multi_elts
        elif data_type == 4 :
            if (create or clone) :
                if nb_elements_per_line == 1 :
                    subclass = OBIDMS_column_char_writable
                else :
                    subclass = OBIDMS_column_char_multi_elts_writable
            else :
                if nb_elements_per_line == 1 :
                    subclass = OBIDMS_column_char
                else :
                    subclass = OBIDMS_column_char_multi_elts
        elif data_type == 5 :
            if (create or clone) :
                if nb_elements_per_line == 1 :
                    subclass = OBIDMS_column_str_writable
                else :
                    subclass = OBIDMS_column_str_multi_elts_writable
            else :
                if nb_elements_per_line == 1 :
                    subclass = OBIDMS_column_str
                else :
                    subclass = OBIDMS_column_str_multi_elts
        else :
            raise Exception("Problem with the data type")
            
        column = subclass(self, column_name, 
                          create, clone, clone_data, 
                          version_number, data_type, 
                          nb_lines, nb_elements_per_line, 
                          elements_names, array_name)

        return column



cdef class OBIDMS_column :

    # Should only be initialized through a subclass
    def __init__(self, 
                 OBIDMS dms, 
                 str column_name,
                 bint create, 
                 bint clone, bint clone_data,
                 obiversion_t version_number, 
                 OBIType_t type,
                 index_t nb_lines, 
                 index_t nb_elements_per_line, 
                 list elements_names,
                 str array_name):
        
        # Declarations
        cdef bytes column_name_b
        cdef bytes dms_name_b
        cdef bytes array_name_b
        cdef bytes elements_names_b
        
        # Fill structure
        self.dms = dms
        self.data_type = bytes2str(name_data_type(type))
        self.column_name = column_name
        self.nb_elements_per_line = nb_elements_per_line
        self.elements_names = elements_names
        
        # Format the character strings to send them to C functions
        column_name_b = str2bytes(column_name)
        dms_name_b = str2bytes(self.dms.dms_name)
        array_name_b = str2bytes(array_name)
        
        # Create, clone or open column
        if create :
            if elements_names == None :  
                elements_names_b = column_name_b
            else :
                elements_names_b = str2bytes(";".join(elements_names))
            self.pointer = obi_create_column(self.dms.pointer, column_name_b, type, 
                                             nb_lines, nb_elements_per_line, 
                                             elements_names_b, array_name_b)
        else :
            if clone :
                self.pointer = obi_clone_column(self.dms.pointer, column_name_b, version_number, clone_data)
            else :
                self.pointer = obi_open_column(self.dms.pointer, column_name_b, version_number)
        
    
    def __iter__(self):
    
        # Declarations
        cdef index_t lines_used
        cdef index_t line_nb
        
        # Yield each line
        lines_used = self.pointer.header.lines_used
        for line_nb in range(lines_used):
            yield self.get_line(line_nb)


    def __setitem__(self, index_t line_nb, object value):
        self.set_line(line_nb, value)


    def __getitem__(self, index_t line_nb):
        return self.get_line(line_nb)


#    cpdef object get_item(self, index_t line_nb, str element_name):           TODO
#        raise NotImplementedError


#    cpdef set_item(self, index_t line_nb, str element_name, object value):    TODO
#        raise NotImplementedError

    
    cpdef list get_elements_names(self):
        return self.elements_names

             
    cpdef str get_data_type(self):
        return self.data_type
    
    
    cpdef index_t get_nb_lines_used(self):
        return self.pointer.header.lines_used
    
    
    cpdef str get_creation_date(self):
        return bytes2str(obi_column_format_date(self.pointer.header.creation_date))
    
    
    cpdef close(self):
        raise NotImplementedError
   
   
