/****************************************************************************
 * OBIDMS array functions	                                                *
 ****************************************************************************/

/**
 * @file obiarray.c
 * @author Celine Mercier
 * @date October 26th 2015
 * @brief Functions handling arrays for storing and retrieving bit arrays.
 */


#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <time.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <math.h>

#include "obiarray.h"
#include "obierrno.h"
#include "obitypes.h"
#include "obidebug.h"
#include "private_at_functions.h"


#define DEBUG_LEVEL 0	// TODO has to be defined somewhere else (cython compil flag?)


/**************************************************************************
 *
 * D E C L A R A T I O N   O F   T H E   P R I V A T E   F U N C T I O N S
 *
 **************************************************************************/

/**
 * @brief Internal function building the file name for an array file.
 *
 * @warning The returned pointer has to be freed by the caller.
 *
 * @param array_name The name of the array.
 *
 * @returns A pointer to the array file name.
 * @retval NULL if an error occurred.
 *
 * @since October 2015
 * @author Celine Mercier (celine.mercier@metabarcoding.org)
 */
static char* build_array_file_name(const char* array_name);


/**
 * @brief Internal function building the file name for an array data file.
 *
 * @warning The returned pointer has to be freed by the caller.
 *
 * @param array_name The name of the array.
 *
 * @returns A pointer to the array data file name.
 * @retval NULL if an error occurred.
 *
 * @since October 2015
 * @author Celine Mercier (celine.mercier@metabarcoding.org)
 */
static char* build_array_data_file_name(const char* array_name);


/**
 * @brief Internal function returning the size of an array header on this platform.
 *
 * @returns The size of an array header in bytes.
 *
 * @since October 2015
 * @author Celine Mercier (celine.mercier@metabarcoding.org)
 */
size_t get_array_header_size();


/**
 * @brief Internal function returning the initial size of an array on this platform.
 *
 * @returns The initial size of an array in bytes.
 *
 * @since October 2015
 * @author Celine Mercier (celine.mercier@metabarcoding.org)
 */
size_t get_initial_array_size();


/**
 * @brief Internal function returning the size of a data array header on this platform.
 *
 * @returns The size of a data array header in bytes.
 *
 * @since October 2015
 * @author Celine Mercier (celine.mercier@metabarcoding.org)
 */
size_t get_array_data_header_size();


/**
 * @brief Internal function returning the initial size of a data array on this platform.
 *
 * @returns The initial size of a data array in bytes.
 *
 * @since October 2015
 * @author Celine Mercier (celine.mercier@metabarcoding.org)
 */
size_t get_initial_array_data_size();


/**
 * @brief Internal function closing a data array structure.
 *
 * @param array_data A pointer to the data array.
 *
 * @retval 0 if the operation was successfully completed.
 * @retval -1 if an error occurred.
 *
 * @since October 2015
 * @author Celine Mercier (celine.mercier@metabarcoding.org)
 */
int close_array_data(OBIDMS_array_data_p array_data);


/**
 * @brief Internal function enlarging an array.
 *
 * @param array A pointer to the array structure.
 *
 * @retval 0 if the operation was successfully completed.
 * @retval -1 if an error occurred.
 *
 * @since October 2015
 * @author Celine Mercier (celine.mercier@metabarcoding.org)
 */
int grow_array(OBIDMS_array_p array);


/**
 * @brief Internal function enlarging a data array.
 *
 * @param array A pointer to the array structure.
 *
 * @retval 0 if the operation was successfully completed.
 * @retval -1 if an error occurred.
 *
 * @since October 2015
 * @author Celine Mercier (celine.mercier@metabarcoding.org)
 */
int grow_array_data(OBIDMS_array_p array);


/**
 * @brief Internal function comparing two byte arrays.
 *
 * The encoding is compared first, then the length of the
 * values, then the values themselves.
 *
 * @param value_1 A pointer to the first byte array.
 * @param value_2 A pointer to the second byte array.
 *
 * @returns A value < 0 if value_1 < value_2,
 * 			a value > 0 if value_1 > value_2,
 * 			and 0 if value_1 == value_2.
 *
 * @since October 2015
 * @author Celine Mercier (celine.mercier@metabarcoding.org)
 */
int array_compare(byte_t* value_1, byte_t* value_2);


/**
 * @brief Internal function calculating the size in bytes of a byte array.
 *
 * @param value A pointer to the byte array.
 *
 * @returns The size of the byte array in bytes.
 *
 * @since October 2015
 * @author Celine Mercier (celine.mercier@metabarcoding.org)
 */
size_t array_sizeof(byte_t* value); 	// TODO not sure about type


/************************************************************************
 *
 * D E F I N I T I O N   O F   T H E   P R I V A T E   F U N C T I O N S
 *
 ************************************************************************/

static char* build_array_file_name(const char* array_name)
{
	char* file_name;

	// Build the file name
	if (asprintf(&file_name,"%s.oda", array_name) < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError building an array file name");
		return NULL;
	}

	// Test if the array name is not too long
	if (strlen(file_name) >= ARRAY_MAX_NAME)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError due to array name too long");
		free(file_name);
		return NULL;
	}

	return file_name;
}


static char* build_array_data_file_name(const char* array_name)
{
	char* file_name;

	// Build the file name
	if (asprintf(&file_name,"%s.odd", array_name) < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError building an array data file name");
		return NULL;
	}

	return file_name;
}


size_t get_array_header_size()
{
	return getpagesize() * 1;
}


size_t get_initial_array_size()
{
	return getpagesize() * 1;
}


size_t get_array_data_header_size()
{
	return getpagesize() * 1;
}


size_t get_initial_array_data_size()
{
	return getpagesize() * 1;
}


int close_array_data(OBIDMS_array_data_p array_data)
{
	int ret_val = 0;

	if (munmap(array_data->header, (array_data->header)->header_size) < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError munmapping the header of an array data file");
		ret_val = -1;
	}

	if (munmap(array_data->data, (array_data->header)->data_size_max) < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError munmapping the data of an array data file");
		ret_val = -1;
	}

	free(array_data);

	return ret_val;
}


int grow_array(OBIDMS_array_p array)		// TODO Lock when needed
{
	size_t file_size;
	size_t old_data_size;
	size_t new_data_size;
	size_t header_size;
	int array_dir_file_descriptor;
	int array_file_descriptor;
	char* array_file_name;

	// Get the file descriptor associated to the array directory
	array_dir_file_descriptor = dirfd(array->directory);
	if (array_dir_file_descriptor < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError getting the file descriptor for an array directory");
		return -1;
	}

	// Get the array file name
	array_file_name = build_array_file_name((array->header)->array_name);
	if (array_file_name == NULL)
	{
		close(array_dir_file_descriptor);
		return -1;
	}

	// Open the array file
	array_file_descriptor = openat(array_dir_file_descriptor, array_file_name, O_RDWR);
	if (array_file_descriptor < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError opening an array file");
		close(array_dir_file_descriptor);
		free(array_file_name);
		return -1;
	}
	free(array_file_name);

	// Calculate the new file size
	old_data_size = (array->header)->array_size;
	new_data_size = old_data_size * ARRAY_GROWTH_FACTOR;
	header_size = (array->header)->header_size;
	file_size = header_size + new_data_size;

	// Enlarge the file // TODO isn't it possible that this makes the file "move"?
	if (ftruncate(array_file_descriptor, file_size) < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError enlarging an array file");
		close(array_dir_file_descriptor);
		close(array_file_descriptor);
		return -1;
	}

	// Unmap and re-map the data

	if (munmap(array->first, old_data_size) < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError munmapping the array of an array file before enlarging");
		close(array_dir_file_descriptor);
		close(array_file_descriptor);
		return -1;
	}

	array->first = mmap(NULL,
					    new_data_size,
					    PROT_READ | PROT_WRITE,
					    MAP_SHARED,
					    array_file_descriptor,
					    header_size
					   );

	if (array->first == MAP_FAILED)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError re-mmapping the array of an array file after enlarging the file");
		close(array_dir_file_descriptor);
		close(array_file_descriptor);
		return -1;
	}

	// Set new maximum number of items
	(array->header)->nb_items_max = ceil(((double) new_data_size) / ((double) sizeof(index_t)));

	// Set the new array size
	(array->header)->array_size = new_data_size;

	// Initialize new data to 0
	memset(((uint8_t*)(array->first))+old_data_size, 0, (new_data_size-old_data_size));

	close(array_file_descriptor);

	return 0;
}


int grow_array_data(OBIDMS_array_p array)		// TODO Lock when needed
{
	size_t file_size;
	index_t old_data_size;
	index_t new_data_size;
	size_t header_size;
	int array_dir_file_descriptor;
	int array_data_file_descriptor;
	char* array_data_file_name;

	// Get the file descriptor associated to the array directory
	array_dir_file_descriptor = dirfd(array->directory);
	if (array_dir_file_descriptor < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError getting the file descriptor for an array directory");
		return -1;
	}

	// Get the array data file name
	array_data_file_name = build_array_data_file_name((array->header)->array_name);
	if (array_data_file_name == NULL)
	{
		close(array_dir_file_descriptor);
		return -1;
	}

	// Open the array data file
	array_data_file_descriptor = openat(array_dir_file_descriptor, array_data_file_name, O_RDWR);
	if (array_data_file_descriptor < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError opening an array data file");
		close(array_dir_file_descriptor);
		free(array_data_file_name);
		return -1;
	}
	free(array_data_file_name);

	// Calculate the new file size
	old_data_size = ((array->data)->header)->data_size_max;
	new_data_size = old_data_size * ARRAY_GROWTH_FACTOR;
	header_size = ((array->data)->header)->header_size;
	file_size = header_size + new_data_size;

	// Enlarge the file // TODO isn't it possible that this makes the file "move"?
	if (ftruncate(array_data_file_descriptor, file_size) < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError enlarging an array data file");
		close(array_dir_file_descriptor);
		close(array_data_file_descriptor);
		return -1;
	}

	// Unmap and re-map the data

	if (munmap((array->data)->data, old_data_size) < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError munmapping the data of an array data file before enlarging");
		close(array_dir_file_descriptor);
		close(array_data_file_descriptor);
		return -1;
	}

	(array->data)->data = mmap(NULL,
							   new_data_size,
							   PROT_READ | PROT_WRITE,
							   MAP_SHARED,
							   array_data_file_descriptor,
							   header_size
							  );

	if ((array->data)->data == MAP_FAILED)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError re-mmapping the data of an array data file after enlarging the file");
		close(array_dir_file_descriptor);
		close(array_data_file_descriptor);
		return -1;
	}

	// Set new data size
	((array->data)->header)->data_size_max = new_data_size;

	// Initialize new data to 0
	memset(((array->data)->data)+old_data_size, 0, new_data_size - old_data_size);

	close(array_data_file_descriptor);

	return 0;
}


int array_compare(byte_t* value_1, byte_t* value_2)
{
	int comp;
	uint8_t size_1;
	uint8_t size_2;
	int32_t len_1;
	int32_t len_2;
	int32_t b;

	//obidebug(1, "\nCOMPARING 1=%d,%.*s; 2=%d,%.*s", *((int32_t*)(value_1+1)), *((int32_t*)(value_1+1)), value_1+BYTE_ARRAY_HEADER_SIZE, *((int32_t*)(value_2+1)), *((int32_t*)(value_2+1)), value_2+BYTE_ARRAY_HEADER_SIZE);

	size_1 = (uint8_t) *(value_1);
	size_2 = (uint8_t) *(value_2);

	if (size_1 != size_2)
		return (size_1 - size_2);

	len_1 = *((int32_t*)(value_1+1));
	len_2 = *((int32_t*)(value_2+1));

	if (len_1 != len_2)
		return (len_1 - len_2);

	b = BYTE_ARRAY_HEADER_SIZE;
	comp = 0;
	while (!comp && (b < len_1+BYTE_ARRAY_HEADER_SIZE))
	{
		comp = *(value_1+b) - *(value_2+b);
		b++;
	}
	return comp;
}


size_t array_sizeof(byte_t* value)
{
	return (BYTE_ARRAY_HEADER_SIZE + *((int32_t*)(value+1)) + 1);
}


/**********************************************************************
 *
 * D E F I N I T I O N   O F   T H E   P U B L I C   F U N C T I O N S
 *
 **********************************************************************/

int obi_array_exists(OBIDMS_p dms, const char* array_name)
{
	struct stat buffer;
	char* array_file_path;
	char* array_file_name;
	int check_dir;
	int array_dir_file_descriptor;

	// Get the file descriptor of the array directory
	array_dir_file_descriptor = dirfd(dms->array_directory);
	if (array_dir_file_descriptor < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError getting the file descriptor for an array directory");
		return -1;
	}

	// Build file name
	array_file_name = build_array_file_name(array_name);
	if (array_file_name == NULL)
	{
		close(array_dir_file_descriptor);
		return -1;
	}

	// Build the array file path
	array_file_path = get_full_path(array_dir_file_descriptor, array_file_name);
	if (array_file_path == NULL)
	{
		obidebug(1, "\nError getting the file path for an array file");
		return -1;
	}

	check_dir = stat(array_file_path, &buffer);

	free(array_file_path);
	free(array_file_name);

	if (check_dir == 0)
		return 1;
	else
		return 0;
}


OBIDMS_array_p obi_array(OBIDMS_p dms, const char* array_name)
{
	int exists;

	exists = obi_array_exists(dms, array_name);

	switch (exists)
	{
		case 0:
			return obi_create_array(dms, array_name);
		case 1:
			return obi_open_array(dms, array_name);
	};

	obidebug(1, "\nError checking if an array already exists");
	return NULL;
}


OBIDMS_array_p obi_create_array(OBIDMS_p dms, const char* array_name)
{
	char* 		array_file_name;
	char* 		array_data_file_name;
	size_t				header_size;
	size_t				data_size;
	size_t				file_size;
	int					array_file_descriptor;
	int					array_data_file_descriptor;
	int					array_dir_file_descriptor;
	OBIDMS_array_data_p array_data;
	OBIDMS_array_p		array;

	// Create the data file

		// Get the file descriptor of the array directory
	array_dir_file_descriptor = dirfd(dms->array_directory);
	if (array_dir_file_descriptor < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError getting the file descriptor for an array directory");
		return NULL;
	}

		// Build file name
	array_data_file_name = build_array_data_file_name(array_name);
	if (array_data_file_name == NULL)
	{
		close(array_dir_file_descriptor);
		return NULL;
	}

		// Create file
	array_data_file_descriptor = openat(array_dir_file_descriptor, array_data_file_name, O_RDWR | O_CREAT | O_EXCL, 0777);
	if (array_data_file_descriptor < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError creating an array data file");
		close(array_dir_file_descriptor);
		free(array_data_file_name);
		return NULL;
	}
	free(array_data_file_name);

		// Calculate the size needed
	header_size = get_array_data_header_size();
	data_size   = get_initial_array_data_size();
	file_size   = header_size + data_size;

		// Truncate the array data file to the right size
	if (ftruncate(array_data_file_descriptor, file_size) < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError truncating an array data file to the right size");
		close(array_dir_file_descriptor);
		close(array_data_file_descriptor);
		return NULL;
	}

		// Allocate the memory for the array data structure
	array_data = (OBIDMS_array_data_p) malloc(sizeof(OBIDMS_array_data_t));
	if (array_data == NULL)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError allocating the memory for the array data structure");
		close(array_dir_file_descriptor);
		close(array_data_file_descriptor);
		return NULL;
	}

		// Fill the array data structure
	array_data->header = mmap(NULL,
			                  header_size,
			                  PROT_READ | PROT_WRITE,
			                  MAP_SHARED,
			                  array_data_file_descriptor,
			                  0
			                 );
	if (array_data->header == MAP_FAILED)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError mmapping the header of an array data file");
		close(array_dir_file_descriptor);
		close(array_data_file_descriptor);
		free(array_data);
		return NULL;
	}

	array_data->data   = mmap(NULL,
			                  data_size,
			                  PROT_READ | PROT_WRITE,
			                  MAP_SHARED,
			                  array_data_file_descriptor,
			                  header_size
			                 );
	if (array_data->data == MAP_FAILED)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError mmapping the data of an array data file");
		munmap(array_data->header, header_size);
		close(array_dir_file_descriptor);
		close(array_data_file_descriptor);
		free(array_data);
		return NULL;
	}

	(array_data->header)->header_size 	 = header_size;
	(array_data->header)->data_size_max	 = data_size;
	(array_data->header)->data_size_used = 0;
	(array_data->header)->nb_items	 	 = 0;
	(array_data->header)->creation_date  = time(NULL);
	strcpy((array_data->header)->array_name, array_name);

	// Initialize all bits to 0
	memset(array_data->data, 0, (array_data->header)->data_size_max);

	close(array_data_file_descriptor);


	// Create the array file

		// Build file name
	array_file_name = build_array_file_name(array_name);
	if (array_file_name == NULL)
	{
		close(array_dir_file_descriptor);
		close_array_data(array_data);
		return NULL;
	}

		// Calculate the size needed
	header_size = get_array_header_size();
	data_size = get_initial_array_size();
	file_size = header_size + data_size;

		// Create file
	array_file_descriptor = openat(array_dir_file_descriptor, array_file_name, O_RDWR | O_CREAT | O_EXCL, 0777);
	if (array_file_descriptor < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError creating an array file");
		close(array_dir_file_descriptor);
		close_array_data(array_data);
		free(array_file_name);
		return NULL;
	}
	free(array_file_name);

		// Truncate the array file to the right size
	if (ftruncate(array_file_descriptor, file_size) < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError truncating an array file to the right size");
		close(array_dir_file_descriptor);
		close_array_data(array_data);
		close(array_file_descriptor);
		return NULL;
	}

		// Allocate the memory for the array structure
	array = (OBIDMS_array_p) malloc(sizeof(OBIDMS_array_t));
	if (array == NULL)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError allocating the memory for the array structure");
		close(array_dir_file_descriptor);
		close_array_data(array_data);
		close(array_file_descriptor);
		return NULL;
	}

		// Fill the array structure
	array->header = mmap(NULL,
			             header_size,
			             PROT_READ | PROT_WRITE,
			             MAP_SHARED,
			             array_file_descriptor,
			             0
			            );
	if (array->header == MAP_FAILED)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError mmapping the header of an array file");
		close(array_dir_file_descriptor);
		close_array_data(array_data);
		close(array_file_descriptor);
		free(array);
		return NULL;
	}

	array->first  = mmap(NULL,
			             data_size,
			             PROT_READ | PROT_WRITE,
			             MAP_SHARED,
			             array_file_descriptor,
			             header_size
			            );
	if (array->first == MAP_FAILED)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError mmapping the data of an array file");
		close(array_dir_file_descriptor);
		close_array_data(array_data);
		munmap(array->header, header_size);
		close(array_file_descriptor);
		free(array);
		return NULL;
	}

	array->data = array_data;
	array->directory = dms->array_directory;

	(array->header)->header_size 	= header_size;
	(array->header)->array_size 	= data_size;
	(array->header)->nb_items	 	= 0;
	(array->header)->nb_items_max	= (index_t) ceil(((double) get_initial_array_size()) / ((double) sizeof(index_t)));
	(array->header)->creation_date 	= time(NULL);
	strcpy((array->header)->array_name, array_name);

	// Initialize to 0	 (TODO: unnecessary?)
	memset(array->first, 0, data_size);

	close(array_file_descriptor);

	return array;
}


OBIDMS_array_p obi_open_array(OBIDMS_p dms, const char* array_name)
{
	char* 	 			array_file_name;
	char* 				array_data_file_name;
	size_t				header_size;
	int					array_file_descriptor;
	int					array_data_file_descriptor;
	int					array_dir_file_descriptor;
	OBIDMS_array_data_p array_data;
	OBIDMS_array_p		array;

	// Open the data file

		// Get the file descriptor of the array directory
	array_dir_file_descriptor = dirfd(dms->array_directory);
	if (array_dir_file_descriptor < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError getting the file descriptor for an array directory");
		return NULL;
	}

		// Build file name
	array_data_file_name = build_array_data_file_name(array_name);
	if (array_data_file_name == NULL)
	{
		close(array_dir_file_descriptor);
		return NULL;
	}

		// Open file
	array_data_file_descriptor = openat(array_dir_file_descriptor, array_data_file_name, O_RDWR, 0777);
	if (array_data_file_descriptor < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError opening an array data file");
		close(array_dir_file_descriptor);
		free(array_data_file_name);
		return NULL;
	}
	free(array_data_file_name);

		// Allocate the memory for the array data structure
	array_data = (OBIDMS_array_data_p) malloc(sizeof(OBIDMS_array_data_t));
	if (array_data == NULL)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError allocating the memory for the array data structure");
		close(array_dir_file_descriptor);
		close(array_data_file_descriptor);
		return NULL;
	}

		// Fill the array data structure
	header_size = get_array_data_header_size();	// TODO should be read?
	array_data->header = mmap(NULL,
			                  header_size,
			                  PROT_READ | PROT_WRITE,
			                  MAP_SHARED,
			                  array_data_file_descriptor,
			                  0
			                 );
	if (array_data->header == MAP_FAILED)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError mmapping the header of an array data file");
		close(array_dir_file_descriptor);
		close(array_data_file_descriptor);
		free(array_data);
		return NULL;
	}

	array_data->data   = mmap(NULL,
			                  (array_data->header)->data_size_max,
			                  PROT_READ | PROT_WRITE,
			                  MAP_SHARED,
			                  array_data_file_descriptor,
			                  header_size
			                 );
	if (array_data->data == MAP_FAILED)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError mmapping the data of an array data file");
		close(array_dir_file_descriptor);
		munmap(array_data->header, header_size);
		close(array_data_file_descriptor);
		free(array_data);
		return NULL;
	}

	close(array_data_file_descriptor);


	// Open the array file

		// Build file name
	array_file_name = build_array_file_name(array_name);
	if (array_file_name == NULL)
	{
		close(array_dir_file_descriptor);
		close_array_data(array_data);
		return NULL;
	}

		// Open file
	array_file_descriptor = openat(array_dir_file_descriptor, array_file_name, O_RDWR, 0777);
	if (array_file_descriptor < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError opening an array file");
		close(array_dir_file_descriptor);
		close_array_data(array_data);
		free(array_file_name);
		return NULL;
	}
	free(array_file_name);

		// Allocate the memory for the array structure
	array = (OBIDMS_array_p) malloc(sizeof(OBIDMS_array_t));
	if (array == NULL)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError allocating the memory for the array structure");
		close(array_dir_file_descriptor);
		close_array_data(array_data);
		close(array_file_descriptor);
		return NULL;
	}

		// Fill the array structure
	header_size = get_array_header_size();	// TODO should be read?
	array->header = mmap(NULL,
			             header_size,
			             PROT_READ | PROT_WRITE,
			             MAP_SHARED,
			             array_file_descriptor,
			             0
			            );
	if (array->header == MAP_FAILED)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError mmapping the header of an array file");
		close(array_dir_file_descriptor);
		close_array_data(array_data);
		close(array_file_descriptor);
		free(array);
		return NULL;
	}

	array->first = mmap(NULL,
						(((array->header)->nb_items_max) * sizeof(index_t)),
			            PROT_READ | PROT_WRITE,
			            MAP_SHARED,
			            array_file_descriptor,
			            header_size
			           );
	if (array->first == MAP_FAILED)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError mmapping the data of an array file");
		close(array_dir_file_descriptor);
		close_array_data(array_data);
		munmap(array->header, header_size);
		close(array_file_descriptor);
		free(array);
		return NULL;
	}

	array->data = array_data;
	array->directory = dms->array_directory;

	close(array_file_descriptor);

	return array;
}


int obi_close_array(OBIDMS_array_p array)
{
	int ret_val = 0;

	ret_val = close_array_data(array->data);

	if (munmap(array->header, (array->header)->header_size) < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError munmapping the header of an array file");
		ret_val = -1;
	}

	if (munmap(array->first, (((array->header)->nb_items_max) * sizeof(index_t))) < 0)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError munmapping the array of an array file");
		ret_val = -1;
	}

	free(array);

	return ret_val;
}


index_t obi_array_add(OBIDMS_array_p array, byte_t* value)
{
	index_t  idx;
	index_t  nb_items;
	index_t  nb_items_max;
	index_t  data_size_used;
	index_t* data_size_max_p;
	size_t   value_size;

	nb_items = (array->header)->nb_items;
	nb_items_max = (array->header)->nb_items_max;
	data_size_used = ((array->data)->header)->data_size_used;
	data_size_max_p = &(((array->data)->header)->data_size_max);

	// Check if the value is already in the array
	if (nb_items > 0)
	{
		idx = obi_array_search(array, value);
		if (idx >= 0)	// The value is already in the array
			return idx;
		else
			idx = -(idx+1);
	}
	else	// First item
		idx = 0;

	// Grow the array if needed
	if (nb_items == nb_items_max)
	{
		if (grow_array(array) < 0)
			return -1;
	}

	// Grow the data if needed
	value_size = array_sizeof(value);
	while (*data_size_max_p < (data_size_used + value_size))	// TODO problem with types
	{
		if (grow_array_data(array) < 0)
			return -1;
	}

	// Store the index of the value in the array

	if (nb_items > 0)
		memmove(((array->first)+idx+1), ((array->first)+idx), ((nb_items - idx) * sizeof(index_t)));
	(array->first)[idx] = data_size_used;

	// Store the value itself at the end of the data
	memcpy((((array->data)->data)+data_size_used), value, value_size);

	// Update the data size
	((array->data)->header)->data_size_used = data_size_used + value_size;

	// Update the number of items
	((array->header)->nb_items)++;
	(((array->data)->header)->nb_items)++;

	// PRINT ARRAY
//	fprintf(stderr, "\nARRAY:");
//	int i;
//	for (i=0; i<=nb_items; i++)
//		fprintf(stderr, "\narray[%d] = %d -> %s", i, a[i], data+a[i]+BYTE_ARRAY_HEADER_SIZE);

	// Return the data index of the value
	return data_size_used;
}


byte_t* obi_array_get(OBIDMS_array_p array, index_t idx)
{
	return (((array->data)->data)+idx);
}


index_t obi_array_search(OBIDMS_array_p array, byte_t* value)
{
	index_t idx_min;
	index_t idx_max;
	index_t idx_mid;
	index_t idx;
	index_t* a;
	byte_t* data;
	byte_t* to_compare;
	index_t nb_items;
	int comp;

	nb_items = (array->header)->nb_items;
	data = (array->data)->data;
	a = array->first;

	if (nb_items == 0)
		return -1;

	idx_max = nb_items-1;
	idx_min = 0;

	while (idx_min <= idx_max)
	{
		idx_mid = (idx_min+idx_max)/2;
		to_compare = obi_array_get(array, a[idx_mid]);
		comp = array_compare(to_compare, value);
		if (!comp)
		{
			// Found
			return a[idx_mid];
		}
		else if (comp < 0)
		{
			// Search in upper half
			idx = idx_mid + 1;
			idx_min = idx_mid + 1;
		}
		else
		{
			// Search in lower half
			idx = idx_mid;
			idx_max = idx_mid - 1;
		}
	}

	// Not found: return where the value should be in the array
	return -(idx+1);
}


byte_t* obi_str_to_obibytes(char* value)
{
	byte_t* value_b;
	int32_t length;
	uint8_t size;

	size = 8;
	length = strlen(value);
	value_b = (byte_t*) malloc(length + BYTE_ARRAY_HEADER_SIZE + 1);
	if (value_b == NULL)
	{
		obi_set_errno(OBI_ARRAY_ERROR);
		obidebug(1, "\nError allocating memory for a byte array");
		return NULL;
	}

	*(value_b) = size;

	*((int32_t*)(value_b+1)) = length;

	strcpy(value_b+BYTE_ARRAY_HEADER_SIZE, value);

	return value_b;
}


const char* obi_obibytes_to_str(byte_t* value_b)
{
	const char* value;

	value = value_b+BYTE_ARRAY_HEADER_SIZE;

	return value;
}

